//#define DEBUG

// DHT22
#define SENSOR_DHT22
#define DHTPIN D7     // Digital pin connected to the DHT sensor.  // dht pins: 1=power, 2=data, 3=NC, 4=GND. 10k from data to power needed 

// PIR Sensors HC-SR501 (modified to put out shortest pulse time)
#define SENSOR_PIR
#define PIRPIN D6  //pir sensor needs 5v. output level is 3.3v

//BH1750 Lux Sensor
#define SENSOR_BH1750


//GPIO2 is blue led on wemos_d1


#include "Arduino.h"

#include <Homie.h>
#define FW_NAME "sensoresp"  //gets printed on topic/$fw/name
#define FW_VERSION "1.0.0"  //gets printed on topic/$fw/version


struct sensordata
{
  unsigned long lastreadtime=0;
  unsigned long readdelay=1000*60; //polling delay
  float minchange=0; //send new value if difference to last sent value is greater than this
  float lastsentvalue=0;
  unsigned long lastsent=0;
  unsigned long senddelaymax=1000*60*10; //maximum time until current value is send
};


#ifdef SENSOR_DHT22
  #include <Adafruit_Sensor.h> //required for dht library
  #include <DHT.h>
  DHT dht(DHTPIN,DHT22,11); //default:11

  struct sensordata dataDHT22_temperature; //struct values are changed in setup()
  struct sensordata dataDHT22_humidity; //struct values are changed in setup()
  
#endif

#ifdef SENSOR_BH1750
  #include <Wire.h>
  #include <BH1750.h>
  BH1750 lightMeter(0x23);
  sensordata dataBH1750;
  /*
  unsigned long lastBH1750time=0;
  unsigned long readdelay_BH1750=2000; //polling delay
  unsigned long lastsend_BH1750=0;
  unsigned long senddelaymax_BH1750=1000*60*10; //maximum time until current value is send
  */
#endif


#ifdef SENSOR_PIR
  sensordata dataPIR;
  /*
  unsigned long lastPIRtime=0;
  unsigned long readdelay_PIR=100; //polling delay
  unsigned long lastsend_PIR=0;
  unsigned long senddelaymax_PIR=1000*60*10; //maximum time until current value is send
  */
  bool lastsentvalue_PIR=false;
#endif


// data/homie/config.json hochladen mit platformio run --target uploadfs
// config contains homie device name, mqtt ip and wifi credentials

/*
  float humidity; //[%RH]  DHT
  float temperature; //[deg C]   DHT
  float light; //[Lux]  BH1750
  bool movement //true bei pir output hight, false wenn low   HC12?501?
*/




HomieNode sensorNode("sensors", "Sensors","sensors");  //id, name, type

char tempstring[16]; //for dtostrf


void loopHandler();
void checkESPStatus();

void setup() {
  Serial.begin(115200);
  Serial.println();
  Serial.println("Booting");

  


  #ifdef SENSOR_DHT22
    Serial.println("initializing dht");
    dht.begin();
    dataDHT22_temperature.minchange=0.1;
    dataDHT22_humidity.minchange=1.0;    
  #endif

  #ifdef SENSOR_BH1750
    Serial.println("initializing bh1750");
    Wire.begin();  //initialize i2c. SDA=D2, SCL=D1
    if (lightMeter.begin(BH1750::CONTINUOUS_HIGH_RES_MODE)) {
      Serial.println(F("BH1750 Advanced begin"));
    } else {
      Serial.println(F("Error initialising BH1750"));
    }
    dataBH1750.minchange=1.0;
  #endif


  #ifdef SENSOR_PIR
    pinMode(PIRPIN, INPUT_PULLUP);
  #endif

  Serial.println("connecting..");
  
  Homie_setFirmware(FW_NAME, FW_VERSION);
  Homie_setBrand(FW_NAME);
  Homie.setLoopFunction(loopHandler);


  #ifdef SENSOR_DHT22
    sensorNode.advertise("temperature");
    sensorNode.advertise("humidity");
  #endif

  #ifdef SENSOR_BH1750
    sensorNode.advertise("light");
  #endif


  #ifdef SENSOR_PIR
    sensorNode.advertise("lastsentvalue_PIR");
  #endif
  

  Homie.setup();
  Serial.println("connected"); //wird nicht ausgegeben. keine ahnung warum.
}

void loop() {
  Homie.loop();  
}


#ifdef SENSOR_DHT22
void loop_DHT22()
{
  sensordata d=dataDHT22_temperature;
  bool _changed=false;
  float temperatureDHT = dht.readTemperature();
  float humidityDHT = dht.readHumidity();

  if (millis() >= (d.lastreadtime+d.readdelay)) {
    temperatureDHT = dht.readTemperature();
    if (abs(d.lastsentvalue-temperatureDHT)>=d.minchange){
      _changed=true;
    }
    d.lastreadtime=millis();
  }

  if (_changed || millis() >= (d.lastsent+d.senddelaymax)) {
    Serial.println("Sending DHT22_temperature");
    checkESPStatus();

    if (!(isnan(temperatureDHT) == 1)){ //success
      Homie.getLogger() << "temperature " << ": " << temperatureDHT << endl;
      sensorNode.setProperty("temperature").send(String(temperatureDHT));
      d.lastsentvalue=temperatureDHT;
    }

    d.lastsent=millis();
  }

  //exchange variables
  d=dataDHT22_humidity;
  _changed=false;

  if (millis() >= (d.lastreadtime+d.readdelay)) {
    humidityDHT = dht.readHumidity();
    if (abs(d.lastsentvalue-humidityDHT)>=d.minchange){
      _changed=true;
    }
    dataPIR.lastreadtime=millis();
  }

  if (_changed || millis() >= (d.lastsent+d.senddelaymax)) {
    Serial.println("Sending DHT22_temperature");
    checkESPStatus();

    if (!(isnan(humidityDHT) == 1)){ //success
      Homie.getLogger() << "humidity " << ": " << humidityDHT << endl;
      sensorNode.setProperty("humidity").send(String(humidityDHT));
      d.lastsentvalue=humidityDHT;
    }
    d.lastsent=millis();
  }
}
#endif

#ifdef SENSOR_BH1750
void loop_BH1750()
{
  sensordata d=dataBH1750;
  if (millis() >= (d.lastreadtime+d.readdelay)) 
  {
    Serial.println("Sending");
    checkESPStatus();

    float light = lightMeter.readLightLevel(); // [lux]
    Homie.getLogger() << "light " << ": " << light << endl;
    sensorNode.setProperty("light").send(String(light));

    d.lastreadtime=millis();
  }
}
#endif

#ifdef SENSOR_PIR
void loop_PIR()
{
  sensordata d=dataPIR;
  bool _changed=false;
  if (millis() >= (d.lastreadtime+d.readdelay)) {
     if (digitalRead(PIRPIN) != lastsentvalue_PIR){
       _changed=true;
     }
     d.lastreadtime=millis();
  }
  if (_changed || millis() >= (d.lastsent+d.senddelaymax)) { //send current value after some long time
    if (digitalRead(PIRPIN)){
       Homie.getLogger() << "lastsentvalue_PIR resend " << ": " << "true" << endl;
       sensorNode.setProperty("lastsentvalue_PIR").send(String("true"));
       lastsentvalue_PIR=true;
     }else{
       Homie.getLogger() << "lastsentvalue_PIR resend " << ": " << "false" << endl;
       sensorNode.setProperty("lastsentvalue_PIR").send(String("false"));
       lastsentvalue_PIR=false;
     }
     d.lastsent=millis();
  }
}
#endif



void loopHandler() {

  #ifdef SENSOR_DHT22
    loop_DHT22();
  #endif

  #ifdef SENSOR_BH1750
    loop_BH1750();
  #endif


  #ifdef SENSOR_PIR
    loop_PIR();
  #endif


  

  

}


void checkESPStatus()
{
  if (WiFi.status() != WL_CONNECTED) //restart if wifi signal loss
  {
    ESP.reset();
  }
}
